//   Copyright 2012 Georg-August-Universität Göttingen, Germany
//
//   Licensed under the Apache License, Version 2.0 (the "License");
//   you may not use this file except in compliance with the License.
//   You may obtain a copy of the License at
//
//       http://www.apache.org/licenses/LICENSE-2.0
//
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and
//   limitations under the License.

package de.ugoe.cs.autoquest.usability.rules.patterns;

import java.util.Arrays;
import java.util.List;

import com.google.common.base.Optional;
import com.google.common.base.Predicate;
import com.google.common.collect.Iterables;

import de.ugoe.cs.autoquest.tasktrees.treeifc.ITaskTree;
import de.ugoe.cs.autoquest.tasktrees.treeifc.ITaskTreeNode;
import de.ugoe.cs.autoquest.usability.tasktree.IterativeDFSFilterStrategy;
import de.ugoe.cs.autoquest.usability.tasktree.filters.TaskTreeFilter;
import de.ugoe.cs.autoquest.usability.tasktree.filters.TaskTreeNodeTypeFilter;

/**
 * <p>
 * TODO comment
 * </p>
 * 
 * @author Alexander Deicke
 */
public class UsagePattern {
    
    private TaskTreeFilter taskTreeFilter = new TaskTreeFilter(new IterativeDFSFilterStrategy());
    
    private TaskTreeNodeTypeFilter concernedNode;

    private List<UsagePatternVisitor> patternVisitors;
    
    private boolean present = false;
    
    /**
     * <p>
     * TODO: comment
     * </p>
     *
     * @param concernedNode
     * @param eventType
     */
    public UsagePattern(TaskTreeNodeTypeFilter concernedNode,
                        UsagePatternVisitor... patternVisitor)
    {
        this.patternVisitors = Arrays.asList(patternVisitor);
        this.concernedNode = concernedNode;
    }

    public boolean containedIn(ITaskTree taskTree) {
        List<ITaskTreeNode> allConcernedNodes = filterAllConcernedNodeFrom(taskTree);
        for(ITaskTreeNode concernedNode : allConcernedNodes) {
            applyAllVisitors(concernedNode);
            if(allVisitorsArePresent()) {
                this.present = true;
                break;
            } else {
                resetAllVisitors();
            }
            
        }
        return this.present;
    }

    private void applyAllVisitors(ITaskTreeNode concernedNode) {
        Optional<UsagePatternVisitor> previousVisitor = Optional.absent();
        for(UsagePatternVisitor visitor : patternVisitors) {
            if(appliedOnSelectionNode(previousVisitor)) {
                for(ITaskTreeNode selectionNode : previousVisitor.get().getRetainedSelectionNodes()) {
                    selectionNode.accept(visitor);
                }
            } else {
                previousVisitor = Optional.of(visitor);
                concernedNode.accept(visitor);
            }
        }
    }

    private boolean appliedOnSelectionNode(Optional<UsagePatternVisitor> previousVisitor) {
        return previousVisitor.isPresent() && previousVisitor.get().hasExcludedSelectionNodes();
    }

    /**
     * <p>
     * TODO: comment
     * </p>
     *
     * @param taskTree
     * @return
     */
    private List<ITaskTreeNode> filterAllConcernedNodeFrom(ITaskTree taskTree) {
        return this.taskTreeFilter.filterByNodeType(this.concernedNode).from(taskTree).nodesMatchedFilter();
    }
    
    /**
     * <p>
     * TODO: comment
     * </p>
     *
     * @return
     */
    private boolean allVisitorsArePresent() {
        Iterable<UsagePatternVisitor> allPresent = Iterables.filter(this.patternVisitors, new Predicate<UsagePatternVisitor>() {
            
            public boolean apply(UsagePatternVisitor visitor) {
                return visitor.isPresent();
            }
            
        });
        return Iterables.size(allPresent) == this.patternVisitors.size();
    }
    
    /**
     * <p>
     * TODO: comment
     * </p>
     *
     */
    private void resetAllVisitors() {
        for(UsagePatternVisitor visitor : this.patternVisitors) {
            visitor.reset();
        }
        
    }

}
